from fluid_sbom.pkg.algorithm import (
    algorithm_length,
)


def infer_algorithm(digest_value: str | None) -> str | None:
    if digest_value:
        for algorithm, length in algorithm_length.items():
            if len(digest_value) == int(length):
                return algorithm.value
    return None



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.model.core import (
    Language,
    Package,
    PackageType,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.python.package import (
    package_url,
)
import logging
from pydantic import (
    ValidationError,
)
from typing import (
    cast,
    ItemsView,
)

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_version(value: IndexedDict) -> str:
    version: str = value["version"]
    return version.strip("=<>~^ ")


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict | None,
) -> list[Package]:
    if dependencies is None:
        return []

    packages = []

    items: ItemsView[str, IndexedDict] = dependencies.items()
    for name, value in items:
        version = _get_version(value)
        if not name or not version:
            continue

        location = _get_location(reader.location, value.position.start.line)

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.PYTHON,
                    type=PackageType.PythonPkg,
                    p_url=package_url(
                        name=name,
                        version=version,
                        package=None,
                    ),
                    licenses=[],
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": location.path(),
                    }
                },
            )
            continue

    return packages


def parse_pipfile_lock_deps(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
    )
    deps: IndexedDict | None = content.get("default")
    dev_deps: IndexedDict | None = content.get("develop")
    packages = [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps),
    ]
    return packages, []



from fluid_sbom.pkg.cataloger.generic.cataloger import (
    Request,
)
from fluid_sbom.pkg.cataloger.javascript.parse_html_scripts import (
    parse_html_scripts,
)
from fluid_sbom.pkg.cataloger.javascript.parse_package_json import (
    parse_package_json,
)
from fluid_sbom.pkg.cataloger.javascript.parse_package_lock import (
    parse_package_lock,
)
from fluid_sbom.pkg.cataloger.javascript.parse_pnpm_lock import (
    parse_pnpm_lock,
)
from fluid_sbom.pkg.cataloger.javascript.parse_yarn_lock import (
    parse_yarn_lock,
)
from fnmatch import (
    fnmatch,
)
import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)


def on_next_javascript(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, x)
                    for x in ("**/package.json", "package.json")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_package_json,
                            parser_name="javascript-parse-package-json",
                        )
                    )
                elif any(
                    fnmatch(value, x)
                    for x in ("**/package-lock.json", "package-lock.json")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_package_lock,
                            parser_name="javascript-parse-package-lock",
                        )
                    )
                elif any(
                    fnmatch(value, x) for x in ("**/yarn.lock", "yarn.lock")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_yarn_lock,
                            parser_name="javascript-parse-yarn-lock",
                        )
                    )
                elif any(
                    fnmatch(value, x)
                    for x in ("**/pnpm-lock.yaml", "pnpm-lock.yaml")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_pnpm_lock,
                            parser_name="javascript-parse-pnpm-lock",
                        )
                    )
                elif fnmatch(value, "*.html"):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_html_scripts,
                            parser_name="javascript-parse-html-scripts",
                        )
                    )
            except Exception as ex:  # pylint:disable=broad-exception-caught
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from .package import (
    package_url,
)
from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.model.core import (
    Language,
    Package,
    PackageType,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
import re

MIX_DEP: re.Pattern[str] = re.compile(
    r"\{:(?P<dep>[\w]*),\s\"~>\s(?P<version>[\d.]+)\".+"
)


def parse_mix_exs(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    is_line_deps = False

    for line_number, line in enumerate(
        reader.read_closer.read().splitlines(), 1
    ):
        line = line.strip()
        if line == "defp deps do":
            is_line_deps = True
        elif is_line_deps:
            if line == "end":
                break
            if matched := MIX_DEP.match(line):
                is_dev = ":dev" in line
                pkg_name = matched.group("dep")
                pkg_version = matched.group("version")
                location = deepcopy(reader.location)
                if location.coordinates:
                    location.coordinates.line = line_number
                packages.append(
                    Package(
                        name=pkg_name,
                        version=pkg_version,
                        type=PackageType.HexPkg,
                        locations=[location],
                        p_url=package_url(pkg_name, pkg_version),
                        metadata=None,
                        language=Language.ELIXIR,
                        licenses=[],
                        is_dev=is_dev,
                    )
                )
    return packages, relationships



import logging
import os

LOGGER = logging.getLogger(__name__)

CI_COMMIT_REF_NAME = os.environ.get("CI_COMMIT_REF_NAME", "local")
CI_COMMIT_SHA = os.environ.get(
    "CI_COMMIT_SHA", "0000000000000000000000000000000000000000"
)
CI_COMMIT_SHORT_SHA = CI_COMMIT_SHA[0:8]
BASE_DIR = os.path.dirname(os.path.abspath(__file__))



from contextlib import (
    suppress,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb import (
    berkeley,
    sqlite,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.entry import (
    header_import,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.errors import (
    InvalidDBFormat,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.package import (
    get_nevra,
    PackageInfo,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.rpmdb_interface import (
    RpmDBInterface,
)


class RpmDB:  # pylint:disable=too-few-public-methods
    def __init__(self, database: RpmDBInterface) -> None:
        self.database = database

    def list_packages(
        self,
    ) -> list[PackageInfo]:
        packages: list[PackageInfo] = []
        for entry in self.database.read():
            try:
                index_entries = header_import(entry)
            except ValueError as exc:
                raise ValueError("Failed to import header") from exc
            if index_entries:
                package = get_nevra(index_entries)
                packages.append(package)
        return packages


def open_db(file_path: str) -> RpmDB | None:
    """
    Attempts to open an RPM database from the specified file path and returns
    an RpmDB instance. If the database is invalid or the metadata cannot be
    validated, None is returned.

    The function first tries to open the database as an SQLite database, and
    if that fails, it attempts to open it as a Berkeley DB. If both attempts
    fail, None is returned.

    :param file_path: The path to the RPM database file.
    :type file_path: str
    :return: An RpmDB instance if the database is valid, otherwise None.
    :rtype: RpmDB | None
    """

    with suppress(InvalidDBFormat):
        return RpmDB(sqlite.open_sqlite(file_path))

    with suppress(InvalidDBFormat):
        return RpmDB(berkeley.open_berkeley(file_path))

    return None



from pydantic import (
    BaseModel,
)


class DotnetDepsEntry(BaseModel):
    name: str
    version: str
    path: str
    sha512: str
    hash_path: str


class DotnetPortableExecutableEntry(BaseModel):
    assembly_version: str | None = None
    legal_copyright: str | None = None
    company_name: str | None = None
    product_name: str | None = None
    product_version: str | None = None
    comments: str | None = None
    internal_name: str | None = None



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.model.core import (
    Package,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.swift.package import (
    new_swift_package_manager_package,
)
import logging
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def parse_package_resolved(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    try:
        package_resolved: IndexedDict = cast(
            IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
        )
    except ValueError:
        return [], []

    packages: list[Package] = []
    relationships: list[Relationship] = []
    package_resolved_pins = package_resolved.get("pins")

    if package_resolved_pins:
        for pin in package_resolved_pins:
            state = pin.get("state", {})
            name = pin.get("identity")
            version = state.get("version")

            if not name or not version:
                continue

            new_location = deepcopy(reader.location)
            if new_location.coordinates:
                new_location.coordinates.line = pin.position.start.line

            if pkg := new_swift_package_manager_package(
                name=name,
                version=version,
                source_url=pin.get("location"),
                revision=state.get("revision"),
                location=new_location,
            ):
                packages.append(pkg)

    return packages, relationships



from datetime import (
    datetime,
)
from fluid_sbom.internal.package_information.go import (
    fetch_latest_version_info,
    fetch_license_info,
)
from fluid_sbom.model.core import (
    HealthMetadata,
    Package,
)
import pydantic


class GolangModuleEntry(pydantic.BaseModel):
    h1_digest: str


def complete_package(package: Package) -> Package:
    latest = fetch_latest_version_info(package.name)
    if not latest:
        return package
    package.health_metadata = HealthMetadata(
        latest_version=latest["Version"],
        latest_version_created_at=datetime.fromisoformat(latest["Time"]),
        artifact=None,
    )
    if package.name.startswith("github.com"):
        license_info = fetch_license_info(
            "/".join(package.name.split("/")[1:])
        )
        if not license_info:
            return package
        licenses = license_info["license"]["spdx_id"]
        package.licenses = [licenses]
    return package
